@@ -15,12 +15,15 @@ import functools
 import logging
 import time
 from typing import (
-    Any, Awaitable, Callable, Iterable, Union,
-    AsyncGenerator, Mapping, Optional, Tuple,
+    Any, Awaitable, Callable, Iterable, Optional, Union,
+    AsyncGenerator,
+    Dict, Mapping,
+    Tuple,
 )
 from urllib.parse import quote as _quote, unquote
 
 from aiotools import aclosing
+from async_timeout import timeout as _timeout
 import etcd3
 from etcd3 import etcdrpc
 from etcd3.client import EtcdTokenCallCredentials
@@ -28,7 +31,7 @@ import grpc
 import trafaret as t
 
 from .logging_utils import BraceStyleAdapter
-from .types import HostPortPair
+from .types import HostPortPair, QueueSentinel
 
 __all__ = (
     'quote', 'unquote',
@@ -39,8 +42,6 @@ Event = namedtuple('Event', 'key event value')
 
 log = BraceStyleAdapter(logging.getLogger(__name__))
 
-sentinel = object()
-
 
 class ConfigScopes(enum.Enum):
     MERGED = 0
@@ -68,13 +69,14 @@ def make_dict_from_pairs(key_prefix, pairs, path_sep='/'):
         path_components = subkey.split('/')
         parent = result
         for p in path_components[:-1]:
+            p = unquote(p)
             if p not in parent:
                 parent[p] = {}
             if p in parent and not isinstance(parent[p], dict):
                 root = parent[p]
                 parent[p] = {'': root}
             parent = parent[p]
-        parent[path_components[-1]] = v
+        parent[unquote(path_components[-1])] = v
     return result
 
 
@@ -117,7 +119,10 @@ def reconn_reauth_adaptor(meth: Callable[..., Awaitable[Any]]):
                 num_reconn_tries += 1
                 continue
             except grpc.RpcError as e:
-                if e.code() == grpc.StatusCode.UNAUTHENTICATED and self._creds:
+                if (
+                    e.code() == grpc.StatusCode.UNAUTHENTICATED or
+                    (e.code() == grpc.StatusCode.UNKNOWN and "invalid auth token" in e.details())
+                ) and self._creds:
                     if num_reauth_tries > 0:
                         raise
                     await reauthenticate(self.etcd_sync, self._creds, self.executor)
@@ -133,13 +138,13 @@ class AsyncEtcd:
 
     def __init__(self, addr: HostPortPair, namespace: str,
                  scope_prefix_map: Mapping[ConfigScopes, str], *,
-                 credentials=None, encoding='utf8', loop=None):
-        self.loop = loop if loop else asyncio.get_event_loop()
+                 credentials=None, encoding='utf8'):
         self.scope_prefix_map = t.Dict({
             t.Key(ConfigScopes.GLOBAL): t.String(allow_blank=True),
             t.Key(ConfigScopes.SGROUP, optional=True): t.String,
             t.Key(ConfigScopes.NODE, optional=True): t.String,
         }).check(scope_prefix_map)
+        self.loop = asyncio.get_running_loop()
         self.executor = ThreadPoolExecutor(max_workers=5, thread_name_prefix='etcd')
         self._creds = credentials
         while True:
@@ -159,6 +164,16 @@ class AsyncEtcd:
         log.info('using etcd cluster from {} with namespace "{}"', addr, namespace)
         self.encoding = encoding
 
+    async def close(self):
+        ret = await self.loop.run_in_executor(
+            self.executor,
+            lambda: self.etcd_sync.close())
+        # ref: https://github.com/kragniz/python-etcd3/issues/997
+        # Currently there is no public API to control this... :(
+        if self.etcd_sync.watcher._callback_thread:
+            self.etcd_sync.watcher._callback_thread.join()
+        return ret
+
     def _mangle_key(self, k: str) -> bytes:
         if k.startswith('/'):
             k = k[1:]
@@ -176,6 +191,15 @@ class AsyncEtcd:
     async def put(self, key: str, val: str, *,
                   scope: ConfigScopes = ConfigScopes.GLOBAL,
                   scope_prefix_map: Mapping[ConfigScopes, str] = None):
+        """
+        Put a single key-value pair to the etcd.
+
+        :param key: The key. This must be quoted by the caller as needed.
+        :param val: The value.
+        :param scope: The config scope for putting the values.
+        :param scope_prefix_map: The scope map used to mangle the prefix for the config scope.
+        :return:
+        """
         scope_prefix_map = ChainMap(scope_prefix_map or {}, self.scope_prefix_map)
         scope_prefix = scope_prefix_map[scope]
         mangled_key = self._mangle_key(f'{_slash(scope_prefix)}{key}')
@@ -183,10 +207,61 @@ class AsyncEtcd:
             self.executor,
             lambda: self.etcd_sync.put(mangled_key, str(val).encode(self.encoding)))
 
+    @reconn_reauth_adaptor
+    async def put_prefix(self, key: str, dict_obj: Mapping[str, str], *,
+                         scope: ConfigScopes = ConfigScopes.GLOBAL,
+                         scope_prefix_map: Mapping[ConfigScopes, str] = None):
+        """
+        Put a nested dict object under the given key prefix.
+        All keys in the dict object are automatically quoted to avoid conflicts with the path separator.
+
+        :param key: Prefix to put the given data. This must be quoted by the caller as needed.
+        :param dict_obj: Nested dictionary representing the data.
+        :param scope: The config scope for putting the values.
+        :param scope_prefix_map: The scope map used to mangle the prefix for the config scope.
+        :return:
+        """
+        scope_prefix_map = ChainMap(scope_prefix_map or {}, self.scope_prefix_map)
+        scope_prefix = scope_prefix_map[scope]
+        flattened_dict: Dict[str, str] = {}
+
+        def _flatten(prefix: str, inner_dict: Mapping[str, str]) -> None:
+            for k, v in inner_dict.items():
+                if k == '':
+                    flattened_key = prefix
+                else:
+                    flattened_key = prefix + '/' + quote(k)
+                if isinstance(v, dict):
+                    _flatten(flattened_key, v)
+                else:
+                    flattened_dict[flattened_key] = v
+
+        _flatten(key, dict_obj)
+
+        return await self.loop.run_in_executor(
+            self.executor,
+            lambda: self.etcd_sync.transaction(
+                [],
+                [self.etcd_sync.transactions.put(
+                    self._mangle_key(f'{_slash(scope_prefix)}{k}'), str(v).encode(self.encoding))
+                    for k, v in flattened_dict.items()],
+                []
+            ))
+
     @reconn_reauth_adaptor
     async def put_dict(self, dict_obj: Mapping[str, str], *,
                        scope: ConfigScopes = ConfigScopes.GLOBAL,
                        scope_prefix_map: Mapping[ConfigScopes, str] = None):
+        """
+        Put a flattened key-value pairs into the etcd.
+        Since the given dict must be a flattened one, its keys must be quoted as needed by the caller.
+        For new codes, ``put_prefix()`` is recommended.
+
+        :param dict_obj: Flattened key-value pairs to put.
+        :param scope: The config scope for putting the values.
+        :param scope_prefix_map: The scope map used to mangle the prefix for the config scope.
+        :return:
+        """
         scope_prefix_map = ChainMap(scope_prefix_map or {}, self.scope_prefix_map)
         scope_prefix = scope_prefix_map[scope]
         return await self.loop.run_in_executor(
@@ -204,6 +279,16 @@ class AsyncEtcd:
                   scope: ConfigScopes = ConfigScopes.MERGED,
                   scope_prefix_map: Mapping[ConfigScopes, str] = None) \
                   -> Optional[str]:
+        """
+        Get a single key from the etcd.
+        Returns ``None`` if the key does not exist.
+        The returned value may be an empty string if the value is a zero-length string.
+
+        :param key: The key. This must be quoted by the caller as needed.
+        :param scope: The config scope to get the value.
+        :param scope_prefix_map: The scope map used to mangle the prefix for the config scope.
+        :return:
+        """
 
         async def get_impl(key: str) -> Optional[str]:
             mangled_key = self._mangle_key(key)
@@ -246,6 +331,40 @@ class AsyncEtcd:
                          scope: ConfigScopes = ConfigScopes.MERGED,
                          scope_prefix_map: Mapping[ConfigScopes, str] = None) \
                          -> Mapping[str, Optional[str]]:
+        """
+        Retrieves all key-value pairs under the given key prefix as a nested dictionary.
+        All dictionary keys are automatically unquoted.
+        If a key has a value while it is also used as path prefix for other keys,
+        the value directly referenced by the key itself is included as a value in a dictionary
+        with the empty-string key.
+
+        For instance, when the etcd database has the following key-value pairs:
+
+        .. code-block::
+
+           myprefix/mydata = abc
+           myprefix/mydata/x = 1
+           myprefix/mydata/y = 2
+           myprefix/mykey = def
+
+        ``get_prefix("myprefix")`` returns the following dictionary:
+
+        .. code-block::
+
+           {
+             "mydata": {
+               "": "abc",
+               "x": "1",
+               "y": "2",
+             },
+             "mykey": "def",
+           }
+
+        :param key_prefix: The key. This must be quoted by the caller as needed.
+        :param scope: The config scope to get the value.
+        :param scope_prefix_map: The scope map used to mangle the prefix for the config scope.
+        :return:
+        """
 
         async def get_prefix_impl(key_prefix: str) -> Iterable[Tuple[str, str]]:
             mangled_key_prefix = self._mangle_key(key_prefix)
@@ -336,27 +455,31 @@ class AsyncEtcd:
             self.executor,
             lambda: self.etcd_sync.delete_prefix(mangled_key_prefix))
 
-    def _watch_cb(self, queue: asyncio.Queue, ev: etcd3.events.Event) -> None:
-        if isinstance(ev, etcd3.events.PutEvent):
-            ev_type = 'put'
-        elif isinstance(ev, etcd3.events.DeleteEvent):
-            ev_type = 'delete'
-        elif isinstance(ev, grpc.RpcError):
-            if ev.code() in (grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.UNKNOWN):
+    def _watch_cb(self, queue: asyncio.Queue, resp: etcd3.watch.WatchResponse) -> None:
+        if isinstance(resp, grpc.RpcError):
+            if (
+                resp.code() == grpc.StatusCode.UNAVAILABLE or
+                (resp.code() == grpc.StatusCode.UNKNOWN and "invalid auth token" not in resp.details())
+            ):
                 # server restarting or terminated
-                self.loop.call_soon_threadsafe(queue.put_nowait, sentinel)
+                self.loop.call_soon_threadsafe(queue.put_nowait, QueueSentinel.CLOSED)
                 return
             else:
-                raise RuntimeError(f'Unexpected RPC Error: {ev}')
-        else:
-            raise TypeError('Not recognized etcd event type.')
-        # etcd3 library uses a separate thread for its watchers.
-        event = Event(
-            self._demangle_key(ev.key),
-            ev_type,
-            ev.value.decode(self.encoding),
-        )
-        self.loop.call_soon_threadsafe(queue.put_nowait, event)
+                raise RuntimeError(f'Unexpected RPC Error: {resp}')
+        for ev in resp.events:
+            if isinstance(ev, etcd3.events.PutEvent):
+                ev_type = 'put'
+            elif isinstance(ev, etcd3.events.DeleteEvent):
+                ev_type = 'delete'
+            else:
+                raise TypeError('Not recognized etcd event type.')
+            # etcd3 library uses a separate thread for its watchers.
+            event = Event(
+                self._demangle_key(ev.key),
+                ev_type,
+                ev.value.decode(self.encoding),
+            )
+            self.loop.call_soon_threadsafe(queue.put_nowait, event)
 
     @reconn_reauth_adaptor
     async def _add_watch_callback(self, raw_key: bytes, cb, **kwargs):
@@ -364,35 +487,61 @@ class AsyncEtcd:
             self.executor,
             lambda: self.etcd_sync.add_watch_callback(raw_key, cb, **kwargs))
 
+    @reconn_reauth_adaptor
+    async def _add_watch_prefix_callback(self, raw_key: bytes, cb, **kwargs):
+        return await self.loop.run_in_executor(
+            self.executor,
+            lambda: self.etcd_sync.add_watch_prefix_callback(raw_key, cb, **kwargs))
+
     @reconn_reauth_adaptor
     async def _cancel_watch(self, watch_id):
         return await self.loop.run_in_executor(
             self.executor,
             lambda: self.etcd_sync.cancel_watch(watch_id))
 
-    async def _watch_impl(self, raw_key: bytes, ready_event: asyncio.Event = None, **kwargs) \
-                         -> AsyncGenerator[Event, None]:
-        queue: asyncio.Queue = asyncio.Queue(loop=self.loop)
+    async def _watch_impl(
+        self,
+        raw_key: bytes,
+        ready_event: asyncio.Event = None,
+        cleanup_event: asyncio.Event = None,
+        prefix: bool = False,
+        timeout: float = None,
+        **kwargs,
+    ) -> AsyncGenerator[Union[QueueSentinel, Event], None]:
+        queue: asyncio.Queue[Union[QueueSentinel, Event]] = asyncio.Queue()
         cb = functools.partial(self._watch_cb, queue)
-        watch_id = await self._add_watch_callback(raw_key, cb, **kwargs)
+        if prefix:
+            watch_id = await self._add_watch_prefix_callback(raw_key, cb, **kwargs)
+        else:
+            watch_id = await self._add_watch_callback(raw_key, cb, **kwargs)
         if ready_event is not None:
             ready_event.set()
         try:
             while True:
-                ev = await queue.get()
-                yield ev
-                queue.task_done()
+                try:
+                    async with _timeout(timeout):
+                        ev = await queue.get()
+                except asyncio.TimeoutError:
+                    yield QueueSentinel.TIMEOUT
+                else:
+                    yield ev
+                    queue.task_done()
         except asyncio.CancelledError:
             raise
         finally:
             await self._cancel_watch(watch_id)
-            del queue
+            if cleanup_event is not None:
+                cleanup_event.set()
 
-    async def watch(self, key: str, *,
-                    scope: ConfigScopes = ConfigScopes.GLOBAL,
-                    scope_prefix_map: Mapping[ConfigScopes, str] = None,
-                    once: bool = False, ready_event: asyncio.Event = None) \
-                    -> AsyncGenerator[Event, None]:
+    async def watch(
+        self, key: str, *,
+        scope: ConfigScopes = ConfigScopes.GLOBAL,
+        scope_prefix_map: Mapping[ConfigScopes, str] = None,
+        once: bool = False,
+        ready_event: asyncio.Event = None,
+        cleanup_event: asyncio.Event = None,
+        wait_timeout: float = None,
+    ) -> AsyncGenerator[Union[QueueSentinel, Event], None]:
         scope_prefix_map = ChainMap(scope_prefix_map or {}, self.scope_prefix_map)
         scope_prefix = scope_prefix_map[scope]
         scope_prefix_len = len(f'{_slash(scope_prefix)}')
@@ -400,36 +549,60 @@ class AsyncEtcd:
         # NOTE: yield from in async-generator is not supported.
         while True:
             try:
-                async with aclosing(self._watch_impl(mangled_key, ready_event)) as agen:
+                async with aclosing(
+                    self._watch_impl(mangled_key,
+                                     ready_event,
+                                     cleanup_event,
+                                     timeout=wait_timeout)
+                ) as agen:
                     async for ev in agen:
-                        if ev is sentinel:
+                        if ev is QueueSentinel.CLOSED:
                             log.debug('watch(): etcd connection closed, restarting watch')
                             break
-                        yield Event(ev.key[scope_prefix_len:], ev.event, ev.value)
-                        if once:
-                            return
+                        elif ev is QueueSentinel.TIMEOUT:
+                            yield ev
+                        else:
+                            yield Event(ev.key[scope_prefix_len:], ev.event, ev.value)
+                            if once:
+                                return
             except asyncio.CancelledError:
-                return
+                break
+        if cleanup_event:
+            cleanup_event.set()
 
-    async def watch_prefix(self, key_prefix: str, *,
-                           scope: ConfigScopes = ConfigScopes.GLOBAL,
-                           scope_prefix_map: Mapping[ConfigScopes, str] = None,
-                           once: bool = False, ready_event: asyncio.Event = None) \
-                           -> AsyncGenerator[Event, None]:
+    async def watch_prefix(
+        self, key_prefix: str, *,
+        scope: ConfigScopes = ConfigScopes.GLOBAL,
+        scope_prefix_map: Mapping[ConfigScopes, str] = None,
+        once: bool = False,
+        ready_event: asyncio.Event = None,
+        cleanup_event: asyncio.Event = None,
+        wait_timeout: float = None,
+    ) -> AsyncGenerator[Union[QueueSentinel, Event], None]:
         scope_prefix_map = ChainMap(scope_prefix_map or {}, self.scope_prefix_map)
         scope_prefix = scope_prefix_map[scope]
         scope_prefix_len = len(f'{_slash(scope_prefix)}')
         mangled_key_prefix = self._mangle_key(f'{_slash(scope_prefix)}{key_prefix}')
-        range_end = etcd3.utils.increment_last_byte(etcd3.utils.to_bytes(mangled_key_prefix))
         while True:
             try:
-                async with aclosing(self._watch_impl(mangled_key_prefix, ready_event, range_end=range_end)) as agen:
+                async with aclosing(
+                    self._watch_impl(mangled_key_prefix,
+                                     ready_event,
+                                     cleanup_event,
+                                     prefix=True,
+                                     timeout=wait_timeout)
+                ) as agen:
                     async for ev in agen:
-                        if ev is sentinel:
+                        if ev is QueueSentinel.CLOSED:
                             log.debug('watch_prefix(): etcd connection closed, restarting watch')
                             break
-                        yield Event(ev.key[scope_prefix_len:], ev.event, ev.value)
-                        if once:
-                            return
+                        elif ev is QueueSentinel.TIMEOUT:
+                            yield ev
+                        else:
+                            yield Event(ev.key[scope_prefix_len:], ev.event, ev.value)
+                            if once:
+                                return
             except asyncio.CancelledError:
-                return
+                break
+        if cleanup_event:
+            cleanup_event.set()
