@@ -1,18 +1,25 @@
 from collections import OrderedDict
+from contextvars import ContextVar
 from datetime import datetime
-import itertools
 import json
 import logging, logging.config, logging.handlers
-import multiprocessing as mp
+import threading
 import os
 from pathlib import Path
-import signal
+import pickle
+import pprint
+from typing import (
+    Any, Optional,
+    Mapping, MutableMapping,
+)
 import socket
 import ssl
+import sys
 
-from setproctitle import setproctitle
 from pythonjsonlogger.jsonlogger import JsonFormatter
 import trafaret as t
+from tblib import pickling_support
+import yarl
 import zmq
 
 from . import config
@@ -20,8 +27,16 @@ from . import validators as tx
 from .logging_utils import BraceStyleAdapter
 from .exception import ConfigurationError
 
-__all__ = ('Logger', 'BraceStyleAdapter')
+# public APIs of this module
+__all__ = (
+    'Logger',
+    'BraceStyleAdapter',
+    'LogstashHandler',
+    'is_active',
+    'pretty',
+)
 
+is_active: ContextVar[bool] = ContextVar('is_active', default=False)
 
 loglevel_iv = t.Enum('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'NOTSET')
 logformat_iv = t.Enum('simple', 'verbose')
@@ -51,6 +66,13 @@ logging_config_iv = t.Dict({
 }).allow_extra('*')
 
 
+class PickledException(Exception):
+    """
+    Serves as a wrapper for exceptions that contain unpicklable arguments.
+    """
+    pass
+
+
 class LogstashHandler(logging.Handler):
 
     def __init__(self, endpoint, protocol: str, *,
@@ -155,14 +177,64 @@ class CustomJsonFormatter(JsonFormatter):
             log_record['level'] = record.levelname
 
 
-def log_worker(daemon_config, parent_pid, log_queue):
-    setproctitle(f'backend.ai: logger pid({parent_pid})')
+class pretty:
+    """A simple object wrapper to pretty-format it when formatting the log record."""
+
+    def __init__(self, obj: Any) -> None:
+        self.obj = obj
+
+    def __repr__(self) -> str:
+        return pprint.pformat(self.obj)
+
+
+def log_worker(
+    daemon_config: Mapping[str, Any],
+    parent_pid: int,
+    log_endpoint: str,
+    ready_event: threading.Event,
+) -> None:
+    console_handler = None
     file_handler = None
     logstash_handler = None
 
+    log_formats = {
+        'simple': '%(levelname)s %(message)s',
+        'verbose': '%(asctime)s %(levelname)s %(name)s [%(process)d] %(message)s',
+    }
+
+    if 'console' in daemon_config['drivers']:
+        drv_config = daemon_config['console']
+        console_formatter: logging.Formatter
+        if drv_config['colored']:
+            import coloredlogs
+            console_formatter = coloredlogs.ColoredFormatter(
+                log_formats[drv_config['format']],
+                field_styles={'levelname': {'color': 248, 'bold': True},
+                              'name': {'color': 246, 'bold': False},
+                              'process': {'color': 'cyan'},
+                              'asctime': {'color': 240}},
+                level_styles={'debug': {'color': 'green'},
+                              'verbose': {'color': 'green', 'bright': True},
+                              'info': {'color': 'cyan', 'bright': True},
+                              'notice': {'color': 'cyan', 'bold': True},
+                              'warning': {'color': 'yellow'},
+                              'error': {'color': 'red', 'bright': True},
+                              'success': {'color': 77},
+                              'critical': {'background': 'red', 'color': 255, 'bold': True}},
+            )
+        else:
+            console_formatter = logging.Formatter(
+                log_formats[drv_config['format']],
+            )
+        console_handler = logging.StreamHandler(
+            stream=sys.stderr,
+        )
+        console_handler.setLevel(daemon_config['level'])
+        console_handler.setFormatter(console_formatter)
+
     if 'file' in daemon_config['drivers']:
         drv_config = daemon_config['file']
-        fmt = '(timestamp) (level) (name) (processName) (message)'
+        fmt = '%(timestamp) %(level) %(name) %(processName) %(message)'
         file_handler = logging.handlers.RotatingFileHandler(
             filename=drv_config['path'] / drv_config['filename'],
             backupCount=drv_config['backup-count'],
@@ -183,23 +255,123 @@ def log_worker(daemon_config, parent_pid, log_queue):
         )
         logstash_handler.setLevel(daemon_config['level'])
 
+    zctx = zmq.Context()
+    agg_sock = zctx.socket(zmq.PULL)
+    agg_sock.bind(log_endpoint)
+    ep_url = yarl.URL(log_endpoint)
+    if ep_url.scheme.lower() == 'ipc':
+        os.chmod(ep_url.path, 0o777)
     try:
+        ready_event.set()
         while True:
-            rec = log_queue.get()
+            data = agg_sock.recv()
+            if not data:
+                return
+            try:
+                rec = pickle.loads(data)
+            except (pickle.PickleError, TypeError):
+                # We have an unpickling error.
+                # Change into a self-created log record with exception info.
+                rec = logging.makeLogRecord({
+                    'name': __name__,
+                    'msg': 'Cannot unpickle the log record (raw data: %r)',
+                    'levelno': logging.ERROR,
+                    'levelname': 'error',
+                    'args': (data,),  # attach the original data for inspection
+                    'exc_info': sys.exc_info(),
+                })
             if rec is None:
                 break
-            if file_handler:
-                file_handler.emit(rec)
-            if logstash_handler:
-                logstash_handler.emit(rec)
+            if console_handler:
+                console_handler.emit(rec)
+            try:
+                if file_handler:
+                    file_handler.emit(rec)
+                if logstash_handler:
+                    logstash_handler.emit(rec)
+            except OSError:
+                # don't terminate the log worker.
+                continue
     finally:
         if logstash_handler:
             logstash_handler.cleanup()
+        agg_sock.close()
+        zctx.term()
+
+
+class RelayHandler(logging.Handler):
+
+    def __init__(self, *, endpoint: str) -> None:
+        super().__init__()
+        self.endpoint = endpoint
+        self._zctx = zmq.Context()
+        # We should use PUSH-PULL socket pairs to avoid
+        # lost of synchronization sentinel messages.
+        if endpoint:
+            self._sock = self._zctx.socket(zmq.PUSH)
+            self._sock.setsockopt(zmq.LINGER, 100)
+            self._sock.connect(self.endpoint)
+        else:
+            self._sock = None
+
+    def close(self) -> None:
+        if self._sock is not None:
+            self._sock.close()
+        self._zctx.term()
+
+    def _fallback(self, record: Optional[logging.LogRecord]) -> None:
+        if record is None:
+            return
+        print(record.getMessage(), file=sys.stderr)
+
+    def emit(self, record: Optional[logging.LogRecord]) -> None:
+        if self._sock is None:
+            self._fallback(record)
+            return
+        # record may be None to signal shutdown.
+        try:
+            if record is not None and record.exc_info is not None:
+                pickling_support.install(record.exc_info[1])
+            pickled_rec = pickle.dumps(record)
+        except (pickle.PickleError, TypeError):
+            # We have a pickling error.
+            # Change it into a self-created picklable log record with exception info.
+            if record is not None:
+                exc_info: Any
+                if isinstance(record.exc_info, tuple):
+                    exc_info = (
+                        PickledException,
+                        PickledException(repr(record.exc_info[1])),  # store stringified repr
+                        record.exc_info[2],
+                    )
+                else:
+                    exc_info = record.exc_info
+                record = logging.makeLogRecord({
+                    'name': record.name,
+                    'pathname': record.pathname,
+                    'lineno': record.lineno,
+                    'msg': record.getMessage(),
+                    'levelno': record.levelno,
+                    'levelname': record.levelname,
+                    'exc_info': exc_info,
+                })
+            pickled_rec = pickle.dumps(record)
+        try:
+            self._sock.send(pickled_rec)
+        except zmq.ZMQError:
+            self._fallback(record)
 
 
 class Logger():
 
-    def __init__(self, daemon_config):
+    is_master: bool
+    log_endpoint: str
+    daemon_config: Mapping[str, Any]
+    log_config: MutableMapping[str, Any]
+    log_worker: threading.Thread
+
+    def __init__(self, daemon_config: MutableMapping[str, Any], *,
+                 is_master: bool, log_endpoint: str) -> None:
         legacy_logfile_path = os.environ.get('BACKEND_LOG_FILE')
         if legacy_logfile_path:
             p = Path(legacy_logfile_path)
@@ -221,16 +393,12 @@ class Logger():
         _check_driver_config_exists_if_activated(cfg, 'file')
         _check_driver_config_exists_if_activated(cfg, 'logstash')
 
+        self.is_master = is_master
+        self.log_endpoint = log_endpoint
         self.daemon_config = cfg
-        self.log_formats = {
-            'simple': '%(levelname)s %(message)s',
-            'verbose': '%(asctime)s %(levelname)s %(name)s [%(process)d] %(message)s',
-        }
         self.log_config = {
             'version': 1,
             'disable_existing_loggers': False,
-            'formatters': {
-            },
             'handlers': {
                 'null': {'class': 'logging.NullHandler'},
             },
@@ -241,69 +409,39 @@ class Logger():
         }
 
     def __enter__(self):
-        self.log_queue = mp.Queue()
-
-        if 'console' in self.daemon_config['drivers']:
-            drv_config = self.daemon_config['console']
-            if drv_config['colored']:
-                self.log_config['formatters']['console'] = {
-                    '()': 'coloredlogs.ColoredFormatter',
-                    'format': self.log_formats[drv_config['format']],
-                    'field_styles': {'levelname': {'color': 248, 'bold': True},
-                                     'name': {'color': 246, 'bold': False},
-                                     'process': {'color': 'cyan'},
-                                     'asctime': {'color': 240}},
-                    'level_styles': {'debug': {'color': 'green'},
-                                     'verbose': {'color': 'green', 'bright': True},
-                                     'info': {'color': 'cyan', 'bright': True},
-                                     'notice': {'color': 'cyan', 'bold': True},
-                                     'warning': {'color': 'yellow'},
-                                     'error': {'color': 'red', 'bright': True},
-                                     'success': {'color': 77},
-                                     'critical': {'background': 'red', 'color': 255, 'bold': True}},
-                }
-            else:
-                self.log_config['formatters']['console'] = {
-                    'class': 'logging.Formatter',
-                    'format': self.log_formats[drv_config['format']],
-                }
-            self.log_config['handlers']['console'] = {
-                'class': 'logging.StreamHandler',
-                'level': self.daemon_config['level'],
-                'formatter': 'console',
-                'stream': 'ext://sys.stderr',
-            }
-            for _logger in self.log_config['loggers'].values():
-                _logger['handlers'].append('console')
-
-        def _activate_aggregator():
-            if 'aggregator' not in self.log_config['handlers']:
-                self.log_config['handlers']['aggregator'] = {
-                    'class': 'logging.handlers.QueueHandler',
-                    'level': self.daemon_config['level'],
-                    'queue': self.log_queue,
-                }
-                for _logger in self.log_config['loggers'].values():
-                    _logger['handlers'].append('aggregator')
-
-        if 'file' in self.daemon_config['drivers']:
-            _activate_aggregator()
-
-        if 'logstash' in self.daemon_config['drivers']:
-            _activate_aggregator()
-
+        if is_active.get():
+            raise RuntimeError('You cannot activate two or more loggers at the same time.')
+        tx.fix_trafaret_pickle_support()  # monkey-patch for pickling trafaret.DataError
+        pickling_support.install()        # enable pickling of tracebacks
+        self.log_config['handlers']['relay'] = {
+            'class': 'ai.backend.common.logging.RelayHandler',
+            'level': self.daemon_config['level'],
+            'endpoint': self.log_endpoint,
+        }
+        for _logger in self.log_config['loggers'].values():
+            _logger['handlers'].append('relay')
         logging.config.dictConfig(self.log_config)
-        # block signals that may interrupt/corrupt logging
-        stop_signals = {signal.SIGINT, signal.SIGTERM}
-        signal.pthread_sigmask(signal.SIG_BLOCK, stop_signals)
-        self.proc = mp.Process(
-            target=log_worker, name='Logger',
-            args=(self.daemon_config, os.getpid(), self.log_queue))
-        self.proc.start()
-        signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
-        # reset process counter
-        mp.process._process_counter = itertools.count(0)
+        self._is_active_token = is_active.set(True)
+        if self.is_master and self.log_endpoint:
+            self.relay_handler = logging.getLogger('').handlers[0]
+            self.ready_event = threading.Event()
+            assert isinstance(self.relay_handler, RelayHandler)
+            self.log_worker = threading.Thread(
+                target=log_worker, name='Logger',
+                args=(self.daemon_config, os.getpid(), self.log_endpoint, self.ready_event))
+            self.log_worker.start()
+            self.ready_event.wait()
 
     def __exit__(self, *exc_info_args):
-        self.log_queue.put(None)
-        self.proc.join()
+        # Resetting generates "different context" errors.
+        # Since practically we only need to check activeness in alembic scripts
+        # and it should be active until the program terminates,
+        # just leave it as-is.
+        is_active.reset(self._is_active_token)
+        if self.is_master and self.log_endpoint:
+            self.relay_handler.emit(None)
+            self.log_worker.join()
+            self.relay_handler.close()
+            ep_url = yarl.URL(self.log_endpoint)
+            if ep_url.scheme.lower() == 'ipc':
+                os.unlink(ep_url.path)
