@@ -1,24 +1,34 @@
+from __future__ import annotations
+
 from collections import UserDict, namedtuple
+from contextvars import ContextVar
 from decimal import Decimal
 import enum
 import ipaddress
 import math
 import numbers
 from typing import (
-    Any, Optional, Union,
-    Tuple, Sequence,
+    Any,
+    List,
+    Literal,
     Mapping,
-    NewType, Type, TypeVar,
+    NewType,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    TypeVar,
+    TypedDict,
+    TYPE_CHECKING,
+    Union,
 )
-from typing_extensions import TypedDict
 import uuid
 
-import attr
-
 __all__ = (
     'aobject',
     'DeviceId',
     'ContainerId',
+    'SessionId',
     'KernelId',
     'MetricKey',
     'MetricValue',
@@ -33,30 +43,35 @@ __all__ = (
     'IntrinsicSlotNames',
     'ResourceSlot',
     'MountPermission',
+    'MountPermissionLiteral',
+    'MountTuple5',
+    'MountTuple4',
+    'MountTuple3',
+    'MountTypes',
     'KernelCreationConfig',
     'KernelCreationResult',
+    'ServicePortProtocols',
 )
 
+if TYPE_CHECKING:
+    from .docker import ImageRef
+
 
 T_aobj = TypeVar('T_aobj', bound='aobject')
 
+current_resource_slots: ContextVar[Mapping[SlotName, SlotTypes]] = ContextVar('current_resource_slots')
+
 
 class aobject(object):
     '''
     An "asynchronous" object which guarantees to invoke both ``def __init__(self, ...)`` and
     ``async def __ainit(self)__`` to ensure asynchronous initialization of the object.
 
-    You can create an instance of subclasses of aboject in two ways:
-
-    .. code-block:: python
-
-       o = await SomeAObj(...)
+    You can create an instance of subclasses of aboject in the following way:
 
     .. code-block:: python
 
        o = await SomeAObj.new(...)
-
-    The latter is supported to avoid type checking errors (currently confirmed in mypy 0.720+).
     '''
 
     @classmethod
@@ -89,8 +104,8 @@ HostPID = NewType('HostPID', PID)
 ContainerPID = NewType('ContainerPID', PID)
 
 ContainerId = NewType('ContainerId', str)
+SessionId = NewType('SessionId', uuid.UUID)
 KernelId = NewType('KernelId', uuid.UUID)
-SessionId = NewType('SessionId', str)
 
 AgentId = NewType('AgentId', str)
 DeviceName = NewType('DeviceName', str)
@@ -105,6 +120,19 @@ SecretKey = NewType('SecretKey', str)
 class SlotTypes(str, enum.Enum):
     COUNT = 'count'
     BYTES = 'bytes'
+    UNIQUE = 'unique'
+
+
+class AutoPullBehavior(str, enum.Enum):
+    DIGEST = 'digest'
+    TAG = 'tag'
+    NONE = 'none'
+
+
+class ServicePortProtocols(str, enum.Enum):
+    HTTP = 'http'
+    TCP = 'tcp'
+    PREOPEN = 'preopen'
 
 
 class SessionTypes(str, enum.Enum):
@@ -166,6 +194,9 @@ class MountPermission(str, enum.Enum):
     RW_DELETE = 'wd'
 
 
+MountPermissionLiteral = Literal['ro', 'rw', 'wd']
+
+
 class MountTypes(str, enum.Enum):
     VOLUME = 'volume'
     BIND = 'bind'
@@ -208,39 +239,66 @@ class BinarySize(int):
     endings = ('ibytes', 'ibyte', 'ib', 'bytes', 'byte', 'b')
 
     @classmethod
-    def from_str(cls, expr):
-        if isinstance(expr, (Decimal, numbers.Integral)):
-            return cls(expr)
-        assert isinstance(expr, str)
-        if expr.lower().startswith('inf'):
+    def _parse_str(cls, expr: str) -> Union[BinarySize, Decimal]:
+        if expr.lower() in ('inf', 'infinite', 'infinity'):
             return Decimal('Infinity')
         orig_expr = expr
         expr = expr.strip().replace('_', '')
         try:
-            return int(expr)
+            return cls(expr)
         except ValueError:
             expr = expr.lower()
+            dec_expr: Decimal
             try:
                 for ending in cls.endings:
                     if expr.endswith(ending):
                         length = len(ending) + 1
                         suffix = expr[-length]
-                        expr = Decimal(expr[:-length])
+                        dec_expr = Decimal(expr[:-length])
                         break
                 else:
-                    # when there is no unit ending (e.g., "2K")
+                    # when there is suffix without scale (e.g., "2K")
                     if not str.isnumeric(expr[-1]):
                         suffix = expr[-1]
-                        expr = Decimal(expr[:-1])
+                        dec_expr = Decimal(expr[:-1])
                     else:
-                        suffix = ' '
+                        # has no suffix and is not an integer
+                        # -> fractional bytes (e.g., 1.5 byte)
+                        raise ValueError('Fractional bytes are not allowed')
             except ArithmeticError:
                 raise ValueError('Unconvertible value', orig_expr)
             try:
                 multiplier = cls.suffix_map[suffix]
-                return cls(expr * multiplier)
             except KeyError:
                 raise ValueError('Unconvertible value', orig_expr)
+            return cls(dec_expr * multiplier)
+
+    @classmethod
+    def finite_from_str(
+        cls,
+        expr: Union[str, Decimal, numbers.Integral],
+    ) -> BinarySize:
+        if isinstance(expr, Decimal):
+            if expr.is_infinite():
+                raise ValueError('infinite values are not allowed')
+            return cls(expr)
+        if isinstance(expr, numbers.Integral):
+            return cls(int(expr))
+        result = cls._parse_str(expr)
+        if isinstance(result, Decimal) and result.is_infinite():
+            raise ValueError('infinite values are not allowed')
+        return cls(int(result))
+
+    @classmethod
+    def from_str(
+        cls,
+        expr: Union[str, Decimal, numbers.Integral],
+    ) -> Union[BinarySize, Decimal]:
+        if isinstance(expr, Decimal):
+            return cls(expr)
+        if isinstance(expr, numbers.Integral):
+            return cls(int(expr))
+        return cls._parse_str(expr)
 
     def _preformat(self):
         scale = self
@@ -299,100 +357,104 @@ class ResourceSlot(UserDict):
 
     __slots__ = ('data', )
 
-    def __init__(self, *args) -> None:
-        super().__init__(*args)
+    def __init__(self, *args, **kwargs) -> None:
+        super().__init__(*args, **kwargs)
 
-    def __add__(self, other):
+    def sync_keys(self, other: ResourceSlot) -> None:
+        self_only_keys = self.data.keys() - other.data.keys()
+        other_only_keys = other.data.keys() - self.data.keys()
+        for k in self_only_keys:
+            other.data[k] = Decimal(0)
+        for k in other_only_keys:
+            self.data[k] = Decimal(0)
+
+    def __add__(self, other: ResourceSlot) -> ResourceSlot:
         assert isinstance(other, ResourceSlot), 'Only can add ResourceSlot to ResourceSlot.'
+        self.sync_keys(other)
         return type(self)({
             k: self.get(k, 0) + other.get(k, 0)
             for k in (self.keys() | other.keys())
         })
 
-    def __sub__(self, other):
+    def __sub__(self, other: ResourceSlot) -> ResourceSlot:
         assert isinstance(other, ResourceSlot), 'Only can subtract ResourceSlot from ResourceSlot.'
-        if other.keys() > self.keys():
-            raise ValueError('Cannot subtract resource slot with more keys!')
+        self.sync_keys(other)
         return type(self)({
             k: self.data[k] - other.get(k, 0)
             for k in self.keys()
         })
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
+        if other is self:
+            return True
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() != other.keys():
-            return False
+        self.sync_keys(other)
         self_values = [self.data[k] for k in sorted(self.data.keys())]
         other_values = [other.data[k] for k in sorted(other.data.keys())]
         return self_values == other_values
 
-    def __ne__(self, other):
+    def __ne__(self, other: object) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() != other.keys():
-            return True
+        self.sync_keys(other)
         return not self.__eq__(other)
 
-    def eq_contains(self, other):
+    def eq_contains(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() < other.keys():
-            raise ValueError('Slots with less keys cannot contain other.')
         common_keys = sorted(other.keys() & self.keys())
+        only_other_keys = other.keys() - self.keys()
         self_values = [self.data[k] for k in common_keys]
         other_values = [other.data[k] for k in common_keys]
-        return self_values == other_values
+        return self_values == other_values and all(other[k] == 0 for k in only_other_keys)
 
-    def eq_contained(self, other):
+    def eq_contained(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() > other.keys():
-            raise ValueError('Slots with more keys cannot be contained in other.')
         common_keys = sorted(other.keys() & self.keys())
+        only_self_keys = self.keys() - other.keys()
         self_values = [self.data[k] for k in common_keys]
         other_values = [other.data[k] for k in common_keys]
-        return self_values == other_values
+        return self_values == other_values and all(self[k] == 0 for k in only_self_keys)
 
-    def __le__(self, other):
+    def __le__(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() > other.keys():
-            raise ValueError('Slots with more keys cannot be smaller than other.')
+        self.sync_keys(other)
         self_values = [self.data[k] for k in self.keys()]
         other_values = [other.data[k] for k in self.keys()]
         return not any(s > o for s, o in zip(self_values, other_values))
 
-    def __lt__(self, other):
+    def __lt__(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() > other.keys():
-            raise ValueError('Slots with more keys cannot be smaller than other.')
+        self.sync_keys(other)
         self_values = [self.data[k] for k in self.keys()]
         other_values = [other.data[k] for k in self.keys()]
         return (not any(s > o for s, o in zip(self_values, other_values)) and
                 not (self_values == other_values))
 
-    def __ge__(self, other):
+    def __ge__(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() < other.keys():
-            raise ValueError('Slots with less keys cannot be larger than other.')
+        self.sync_keys(other)
         self_values = [self.data[k] for k in other.keys()]
         other_values = [other.data[k] for k in other.keys()]
         return not any(s < o for s, o in zip(self_values, other_values))
 
-    def __gt__(self, other):
+    def __gt__(self, other: ResourceSlot) -> bool:
         assert isinstance(other, ResourceSlot), 'Only can compare ResourceSlot objects.'
-        if self.keys() < other.keys():
-            raise ValueError('Slots with less keys cannot be larger than other.')
+        self.sync_keys(other)
         self_values = [self.data[k] for k in other.keys()]
         other_values = [other.data[k] for k in other.keys()]
         return (not any(s < o for s, o in zip(self_values, other_values)) and
                 not (self_values == other_values))
 
-    def filter_slots(self, known_slots) -> 'ResourceSlot':
-        if isinstance(known_slots, Mapping):
-            slots = {*known_slots.keys()}
-        else:
-            slots = {*known_slots}
+    def normalize_slots(self, *, ignore_unknown: bool) -> ResourceSlot:
+        known_slots = current_resource_slots.get()
+        unset_slots = known_slots.keys() - self.data.keys()
+        if not ignore_unknown and (unknown_slots := self.data.keys() - known_slots.keys()):
+            raise ValueError('Unknown slots', unknown_slots)
         data = {
             k: v for k, v in self.data.items()
-            if k in slots
+            if k in known_slots
         }
+        for k in unset_slots:
+            data[k] = Decimal(0)
         return type(self)(data)
 
     @classmethod
@@ -493,14 +555,14 @@ class ResourceSlot(UserDict):
         }
 
 
-@attr.s(auto_attribs=True, slots=True)
-class VFolderRequest:
-    '''
-    Represents vfolders for a new compute session request.
-    '''
-    vfolder_id: str
-    vfolder_host: str
-    permission: MountPermission
+MountTuple5 = Tuple[str, str, str, MountPermissionLiteral, str]  # vfname, vfhost, vfid, perm, hostpath
+MountTuple4 = Tuple[str, str, str, MountPermissionLiteral]       # vfname, vfhost, vfid, perm
+MountTuple3 = Tuple[str, str, str]                               # vfname, vfhost, vfid, "rw"
+MountExpression = Union[
+    MountTuple5,
+    MountTuple4,
+    MountTuple3,
+]
 
 
 class ImageRegistry(TypedDict):
@@ -520,7 +582,7 @@ class ImageConfig(TypedDict):
 
 class ServicePort(TypedDict):
     name: str
-    protocol: str
+    protocol: ServicePortProtocols
     container_ports: Sequence[int]
     host_ports: Sequence[Optional[int]]
 
@@ -528,6 +590,7 @@ class ServicePort(TypedDict):
 class DeviceModelInfo(TypedDict):
     device_id: DeviceId
     model_name: str
+    data: Mapping[str, Any]
 
 
 class KernelCreationResult(TypedDict):
@@ -545,14 +608,29 @@ class KernelCreationResult(TypedDict):
 
 class KernelCreationConfig(TypedDict):
     image: ImageConfig
-    session_type: str                  # value of SessionTypes
+    auto_pull: AutoPullBehavior
+    session_type: SessionTypes
+    cluster: dict
     resource_slots: Mapping[str, str]  # json form of ResourceSlot
     resource_opts: Mapping[str, str]   # json form of resource options
     environ: Mapping[str, str]
-    mounts: Sequence[str]              # list of mount expressions
+    mounts: Sequence[MountExpression]  # list of mount expression tuples
+    mount_map: Mapping[str, str]       # Mapping of vfolder custom mount path
+    package_directory: Sequence[str]
     idle_timeout: int
+    bootstrap_script: Optional[str]
     startup_command: Optional[str]
     internal_data: Optional[Mapping[str, Any]]
+    preopen_ports: List[int]
+
+
+class KernelEnqueueingConfig(TypedDict):
+    image_ref: ImageRef
+    cluster_role: str
+    idx: int
+    creation_config: dict
+    bootstrap_script: str
+    startup_command: str
 
 
 def _stringify_number(v: Union[BinarySize, int, float, Decimal]) -> str:
@@ -575,5 +653,10 @@ def _stringify_number(v: Union[BinarySize, int, float, Decimal]) -> str:
     return result
 
 
-class Sentinel(object):
-    pass
+class Sentinel(enum.Enum):
+    TOKEN = 0
+
+
+class QueueSentinel(enum.Enum):
+    CLOSED = 0
+    TIMEOUT = 1
