@@ -3,25 +3,42 @@ An extension module to Trafaret which provides additional type checkers.
 '''
 
 import datetime
+from decimal import Decimal
 import enum
 import ipaddress
+import json
 import os
-from pathlib import Path as _Path
+from pathlib import (
+    PurePath as _PurePath,
+    Path as _Path,
+)
 import re
 from typing import (
-    Any, Optional,
-    List, Mapping, Sequence, Tuple,
-    TypeVar, Type,
+    Any,
+    List,
+    Literal,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
 )
-from typing_extensions import Literal
 import uuid
 import pwd
 
 import dateutil.tz
+try:
+    import jwt
+    jwt_available = True
+except ImportError:
+    jwt_available = False
 import multidict
 import trafaret as t
 from trafaret.base import TrafaretMeta
 from trafaret.lib import _empty
+import yarl
 
 from .types import (
     BinarySize as _BinarySize,
@@ -32,19 +49,33 @@ __all__ = (
     'AliasedKey',
     'MultiKey',
     'BinarySize',
+    'StringList',
     'Enum',
+    'JSONString',
+    'PurePath',
+    'Path',
     'IPNetwork',
     'IPAddress',
     'HostPortPair',
-    'Path',
     'PortRange',
     'UserID',
     'GroupID',
-    'Slug',
+    'UUID',
     'TimeZone',
+    'TimeDuration',
+    'Slug',
+    'URL',
 )
 
 
+def fix_trafaret_pickle_support():
+
+    def __reduce__(self):
+        return (type(self), (self.error, self.name, self.value, self.trafaret, self.code))
+
+    t.DataError.__reduce__ = __reduce__
+
+
 class StringLengthMeta(TrafaretMeta):
     '''
     A metaclass that makes string-like trafarets to have sliced min/max length indicator.
@@ -111,13 +142,29 @@ class MultiKey(t.Key):
 
 class BinarySize(t.Trafaret):
 
-    def check_and_return(self, value: Any) -> _BinarySize:
+    def check_and_return(self, value: Any) -> Union[_BinarySize, Decimal]:
         try:
+            if not isinstance(value, str):
+                value = str(value)
             return _BinarySize.from_str(value)
         except ValueError:
             self._failure('value is not a valid binary size', value=value)
 
 
+class StringList(t.Trafaret):
+
+    def __init__(self, *, delimiter: str = ',') -> None:
+        self.delimiter = delimiter
+
+    def check_and_return(self, value: Any) -> Sequence[str]:
+        try:
+            if not isinstance(value, str):
+                value = str(value)
+            return value.split(self.delimiter)
+        except ValueError:
+            self._failure('value is not a string or not convertible to string', value=value)
+
+
 T_enum = TypeVar('T_enum', bound=enum.Enum)
 
 
@@ -138,19 +185,57 @@ class Enum(t.Trafaret):
                           value=value)
 
 
-class Path(t.Trafaret):
+class JSONString(t.Trafaret):
 
-    def __init__(self, *, type: Literal['dir', 'file'],
-                 base_path: _Path = None,
-                 auto_create: bool = False,
-                 allow_nonexisting: bool = False,
-                 allow_devnull: bool = False,
-                 resolve: bool = True):
+    def check_and_return(self, value: Any) -> dict:
+        try:
+            return json.loads(value)
+        except (KeyError, ValueError):
+            self._failure('value is not a valid JSON string', value=value)
+
+
+class PurePath(t.Trafaret):
+
+    def __init__(
+        self, *,
+        base_path: _PurePath = None,
+        relative_only: bool = False,
+    ) -> None:
         super().__init__()
+        self._base_path = base_path
+        self._relative_only = relative_only
+
+    def check_and_return(self, value: Any) -> _PurePath:
+        p = _PurePath(value)
+        if self._relative_only and p.is_absolute():
+            self._failure('expected relative path but the value is absolute', value=value)
+        if self._base_path is not None:
+            try:
+                p.relative_to(self._base_path)
+            except ValueError:
+                self._failure('value is not in the base path', value=value)
+        return p
+
+
+class Path(PurePath):
+
+    def __init__(
+        self, *,
+        type: Literal['dir', 'file'],
+        base_path: _Path = None,
+        auto_create: bool = False,
+        allow_nonexisting: bool = False,
+        allow_devnull: bool = False,
+        relative_only: bool = False,
+        resolve: bool = True,
+    ) -> None:
+        super().__init__(
+            base_path=base_path,
+            relative_only=relative_only,
+        )
         self._type = type
         if auto_create and type != 'dir':
             raise TypeError('Only directory paths can be set auto-created.')
-        self._base_path = base_path
         self._auto_create = auto_create
         self._allow_nonexisting = allow_nonexisting
         self._allow_devnull = allow_devnull
@@ -161,9 +246,11 @@ class Path(t.Trafaret):
             p = _Path(value).resolve() if self._resolve else _Path(value)
         except (TypeError, ValueError):
             self._failure('cannot parse value as a path', value=value)
+        if self._relative_only and p.is_absolute():
+            self._failure('expected relative path but the value is absolute', value=value)
         if self._base_path is not None:
             try:
-                _base_path = self._base_path.resolve() if self._resolve else self._base_path
+                _base_path = _Path(self._base_path).resolve() if self._resolve else self._base_path
                 p.relative_to(_base_path)
             except ValueError:
                 self._failure('value is not in the base path', value=value)
@@ -223,13 +310,16 @@ class HostPortPair(t.Trafaret):
         else:
             self._failure('urecognized value type', value=value)
         try:
-            host = ipaddress.ip_address(host.strip('[]'))
+            if isinstance(host, str):
+                host = ipaddress.ip_address(host.strip('[]'))
+            elif isinstance(host, ipaddress._BaseAddress):
+                pass
         except ValueError:
             pass  # just treat as a string hostname
         if not self._allow_blank_host and not host:
             self._failure('value has empty host', value=value)
         try:
-            port = t.Int[1:65535].check(port)
+            port = t.ToInt[1:65535].check(port)
         except t.DataError:
             self._failure('port number must be between 1 and 65535', value=value)
         return _HostPortPair(host, port)
@@ -332,6 +422,8 @@ class UUID(t.Trafaret):
 
     def check_and_return(self, value: Any) -> uuid.UUID:
         try:
+            if isinstance(value, uuid.UUID):
+                return value
             if isinstance(value, str):
                 return uuid.UUID(value)
             elif isinstance(value, bytes):
@@ -353,6 +445,41 @@ class TimeZone(t.Trafaret):
         return tz
 
 
+class TimeDuration(t.Trafaret):
+
+    def __init__(self, *, allow_negative: bool = False):
+        self._allow_negative = allow_negative
+
+    def check_and_return(self, value: Any) -> datetime.timedelta:
+        if not isinstance(value, str):
+            self._failure('value must be string', value=value)
+        if len(value) == 0:
+            self._failure('value must not be empty', value=value)
+        try:
+            unit = value[-1]
+            if unit.isdigit():
+                t = float(value)
+                if not self._allow_negative and t < 0:
+                    self._failure('value must be positive', value=value)
+                return datetime.timedelta(seconds=t)
+            else:
+                t = float(value[:-1])
+                if not self._allow_negative and t < 0:
+                    self._failure('value must be positive', value=value)
+                if value[-1] == 'w':
+                    return datetime.timedelta(weeks=t)
+                elif value[-1] == 'd':
+                    return datetime.timedelta(days=t)
+                elif value[-1] == 'h':
+                    return datetime.timedelta(hours=t)
+                elif value[-1] == 'm':
+                    return datetime.timedelta(minutes=t)
+                else:
+                    self._failure('value is not a known time duration', value=value)
+        except ValueError:
+            self._failure(f'invalid numeric literal: {value[:-1]}', value=value)
+
+
 class Slug(t.Trafaret, metaclass=StringLengthMeta):
 
     _rx_slug = re.compile(r'^[a-zA-Z0-9]([a-zA-Z0-9._-]*[a-zA-Z0-9])?$')
@@ -386,3 +513,52 @@ class Slug(t.Trafaret, metaclass=StringLengthMeta):
         else:
             self._failure('value must be a string', value=value)
         return value
+
+
+if jwt_available:
+    class JsonWebToken(t.Trafaret):
+
+        default_algorithms = ['HS256']
+
+        def __init__(
+            self, *,
+            secret: str,
+            inner_iv: t.Trafaret = None,
+            algorithms: Sequence[str] = default_algorithms,
+        ) -> None:
+            self.secret = secret
+            self.algorithms = algorithms
+            self.inner_iv = inner_iv
+
+        def check_and_return(self, value: Any) -> Mapping[str, Any]:
+            try:
+                token_data = jwt.decode(value, self.secret, alogrithms=self.algorithms)
+                if self.inner_iv is not None:
+                    return self.inner_iv.check(token_data)
+                return token_data
+            except jwt.PyJWTError:
+                self._failure('cannot decode the given value as JWT', value=value)
+
+
+class URL(t.Trafaret):
+
+    rx_scheme = re.compile(r"^[-a-z0-9]+://")
+
+    def __init__(
+        self, *,
+        scheme_required: bool = True,
+    ) -> None:
+        self.scheme_required = scheme_required
+
+    def check_and_return(self, value: Any) -> yarl.URL:
+        if not isinstance(value, (str, bytes)):
+            self._failure("A URL must be a unicode string or a byte sequence", value=value)
+        if isinstance(value, bytes):
+            value = value.decode('utf-8')
+        if self.scheme_required:
+            if not self.rx_scheme.match(value):
+                self._failure("The given value does not have the scheme (protocol) part", value=value)
+        try:
+            return yarl.URL(value)
+        except ValueError as e:
+            self._failure(f"cannot convert the given value to URL (error: {e!r})", value=value)
